QuestHubber = LibStub("AceAddon-3.0"):NewAddon("QuestHubber", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceHook-3.0");
local core = QuestHubber;
local L = LibStub("AceLocale-3.0"):GetLocale("QuestHubber");
local G = LibStub:GetLibrary("LibGratuity-3.0");

--[[local options = {
	name = "QuestGather",
	type = 'group',
	args = {
	},
};]]
local defaults = {
	profile = {
		debug = 0,
		display = true,
	},
};
local FACTION_MAP = {
	Alliance = 1,
	Horde = 2,
};
local IS_ENGLISH = (GetLocale()=="enUS" or GetLocale()=="enGB")
local RACE_ID = {
	["Human"] 		= 1,
	["Orc"] 		= 2,
	["Dwarf"] 		= 4,
	["Night Elf"]	= 8,
	["Scourge"]		= 16,
	["Tauren"]		= 32,
	["Gnome"]		= 64,
	["Troll"]		= 128,
	["Goblin"]		= 256,
	["Blood Elf"]	= 512,
	["Draenei"]		= 1024,
	["Worgen"]		= 2048,
}
local CLASS_ID = {
	["WARRIOR"] 	= 1,
	["PALADIN"] 	= 2,
	["HUNTER"] 		= 4,
	["ROGUE"] 		= 8,
	["PRIEST"] 		= 16,
	["DEATH KNIGHT"]= 32,
	["MAGE"] 		= 128,
	["SHAMAN"] 		= 64,
	["WARLOCK"] 	= 256,
	["DRUID"] 		= 1024,
}

-- TODO (possibly)
-- package scripts (on ignore)
-- change the way unknown quests show (if I have any anymore?)
-- npc scan for objects, if possible

core.failedQueries = 0

core.dataLoaded = false;		-- flag set after LoadStep() is done
core.currentZone = -1;			-- current World Map zone id
core.pinReg = {};				-- registry of all used pins
core.frameReg = {};				-- registry of all unused pin frmaes
core.playerQuests = {};			-- quests the player has completed, queried from server
core.showDaily = false;			-- flag to show daily quests (TODO: implement as option)
core.questLink = {				-- table linking optional quests to the mandatory quests they lead to which take away the initial
	[25985] = 27874,
	[27727] = 27203,
};
core.questLog = {};				-- current quest log, used to see when quests are accepted or completed
core.questInit = false;			-- flag set after the quest log has been initially recorded
core.abandonQuest = -1;			-- when "Abandon" is clicked, sets the quest that was selected at the time; used after confirmation

core.cache = {
	quests = {},		-- [questid] = "Localized Name",
	npcs = {},			-- [npcid] = "Localized Name",
	-- zones = {},			-- [zoneid] = "Localized Name",  -- NOT USED
	invalid = {},		-- [npcid] = true, -- flag set so you don't get double debug warnings about a single npc
	schedule = {},		-- [questid] = true, -- scheduled for rescan due to 
};
core.data = {
	zones = {},
	npcs = {}, -- npcs that drop a quest item
	quests = {}, -- unused
};

function core:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("QuestHubberDB", defaults, "Default");
	self:RegisterChatCommand("qhub", function()
		self:Print(format("Failed retrieving quests %d time(s). Current status: %s.", self.failedQueries, self.serverQueried and "Successful" or "Waiting"))
	end)
	--[[
		LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("", options);
		local ACD = LibStub("AceConfigDialog-3.0");
		ACD:AddToBlizOptions("", "");
		self:SetEnabledState(self.db.profile.enabled;
	]]
	
	self:SetupMapModule();
	self:SetupTooltip();
	
	--[[ start the loader if we have data; otherwise, wait for QuestGather
	if (not self.data or not self.data.quests) then
		self.data = {};
		self.data.quests = {};
	else
		--self:InitializeLoader();
	end]]
	
end

-- unused?
local stepSpace = 0.05 				-- time to spend in seconds
function core:InitializeLoader()
	-- this is called when we have data
	self.loader = CreateFrame("Frame");
	local i = 0;
	self.loader:SetScript("OnUpdate", function(_, elapsed)
		i = i + elapsed;
		if (i > stepSpace) then
			self:LoadStep();
		end
	end);
end

function core:OnEnable()
	self:RegisterEvent("WORLD_MAP_UPDATE"); 			-- map change
	self:RegisterEvent("QUEST_QUERY_COMPLETE"); 		-- get quests completed
	
	self:RegisterEvent("QUEST_LOG_UPDATE");				-- scan for quest completion/pickup
	
	self:RegisterEvent("MINIMAP_UPDATE_TRACKING"); 		-- for toggling Low Level Quests
	
	self:Hook("SetAbandonQuest", true);					-- hook for abandoned quests
	self:Hook("AbandonQuest", true);					-- hook for confirmation
	
	self:QueryQuestsCompleted();
	self.mapModule:Show();
end

function core:OnDisable()
	self.mapModule:Hide();	
end

function core:SlashHandle(message)

end

-- currently unused
local npc = {"begin", "complete"};	-- table for looping through npcs in each quest
local lastStep = nil;				-- iterator position
local stepTime = 0.02; 				-- time to spend in seconds
function core:LoadStep()
	local start = GetTime();
	local steps = 0;
	-- custom iterator that picks up from where the last LoadStep left off
	for q, v in next, self.data.quests, lastStep do
		if (tonumber(q)) then	-- make sure it's a fully cached quest
			-- Load Step
			steps = steps + 1
			
			for i, key in ipairs(npc) do
			
				-- cache npc id -> name
				if (v[key] and v[key]["npcid"]) then
					-- read from tooltip
					local id = v[key]["npcid"];
					-- check cache
					local name = self.cache.npcs[id];
					if (not name) then
						self:NPCScan(id);
					end
					if (name) then
						-- update name
						v[key]["npc"] = name;
					end
				end
				
				--[[ cache zone id -> name
				if (v[key] and v[key]["zoneid"]) then
					local id = v[key]["zoneid"];
					local name = self.cache.zones[id];
					
					if (not name and tonumber(id)) then	
						SetMapByID(id);
						name = GetMapInfo();
					end
					if (name) then
						self.cache.zones[id] = name;
						-- also setting english name -> id cache
						-- if (v[key]["zone"]) then
							-- self.cache.engZones[(v[key]["zone"])] = id;
						-- end 
					end
				end]]
				
				-- set data by zone id
				if (v[key] and v[key]["zoneid"]) then
					local id = v[key]["zoneid"];
					if (id == 301) then
						-- self:Print(q);
					end
					if (not self.data.zones[id]) then
						self.data.zones[id] = {};
					end
					
					self.data.zones[id][q] = v;
				end
			end
			
			if (GetTime()-start >= stepTime) then
				self:Debug("Breaking:", steps, "steps; ", ("%.3fs"):format(GetTime()-start))
				break;
			end
		end
		
		lastStep = q;
	end
	
	
	if (not next(self.data.quests, lastStep)) then
		--[[ temp test
		self.data.zones[301]={} -- Stormwind City
		for i=1,10 do
			self.data.zones[301][i^2] = {
				begin = {
					x = i/10,
					y = i/10,
				},
				name = "Test "..i,
			}
		end]]
		
		self.dataLoaded = true;
		self:Debug("Data loaded");
		self.loader:SetScript("OnUpdate", nil);
		self:QueryQuestsCompleted();
		self:UpdatePins(true);
	end
end

function core:QueryQuestsCompleted()
	QueryQuestsCompleted();
	self:ScheduleTimer(function()
		if (not next(self.playerQuests)) then
			self:Debug("Could not retrieve quest data, trying again.");
			self:QueryQuestsCompleted();
			self.failedQueries = self.failedQueries + 1;
		else
			self.mapModule.note:Hide();
			self.serverQueried = true;
		end
	end, 10);
end

-- Unused?
function core:QueryQuestNames()
	self:Debug("QueryQuestNames START");
	local start = GetTime();
	local loop = 0;
	for i,v in pairs(self.cache.schedule) do
		if (i > 0) then
			loop = loop + 1;
			-- self:QuestScan(id);
			if (self.cache.quests[i]) then 
				self.cache.schedule[i] = nil
			end
		end
		if GetTime()-start > -1 then
			self:Debug(format("Breaking secondary lookup after %d loops; %.3fs", loop, start));
			break;
		end
	end
	
	if (next(self.cache.schedule)) then
		self:ScheduleTimer("QueryQuestNames", 20);
		self:Debug("Some quest names were not cached. Checking in 20s");
	end
	self:Debug("QueryQuestNames END");
end


local numScans = 0
local curScans = {}
function core:QuestScan(id)
	-- self:Debug("Scanning", id)
	if (id and id > 0) then
		-- check if we're scanning too many
		if (numScans < 100 or curScans[id]) then
			G:SetHyperlink(format("quest:%d", id));
			local text = G:GetText();
			if (text and text[1] and text[1][1]) then
				-- success, remove from scan list if it was queried
				if (curScans[id]) then
					curScans[id] = nil;
					numScans = numScans - 1;
					-- self:Debug("Removing a query", numScans);
				end
				self.cache.quests[id] = text[1][1];
				local mouse = GetMouseFocus()
				if (mouse and mouse.questHubberID == id) then
					self:ShowTooltip(mouse)
				end
				-- self:Debug(format("Quest id %d = %q", q, text[1][1]))
			else
				-- self:Debug("No name for quest "..id);
				-- self.cache.schedule[id] = true
				-- no result, so the server has been queried
				if (not curScans[id]) then
					numScans = numScans + 1;
					curScans[id] = true;
					-- self:Debug("Adding another query", numScans);
				end
				self:ScheduleTimer(function()
					self:QuestScan(id)
				end, 2)
			end
		else
			-- self:Debug("NOT QUERYING, TOO HIGH")
			self:ScheduleTimer(function()
				self:QuestScan(id)
			end, 2.5)
		end
	end
end

function core:NPCScan(id)
	G:SetHyperlink(format("unit:0xF53%05X00000000", id));
	local text = G:GetText();
	if not text and not self.cache.invalid[id] then
		-- self:Debug("Invalid NPC id:", id, "- perhaps not cached?");
		self.cache.invalid[id] = true;
	end
	if (text and text[1] and text[1][1]) then
		self.cache.npcs[id] = text[1][1];
		name = text[1][1];
	end
end

function core:WORLD_MAP_UPDATE()
	if (self.dataLoaded) then
		self:UpdatePins();
	end
end

-----------------------------------------------
-- Pin management
-----------------------------------------------

function core:UpdatePins(force)
	local id = GetCurrentMapAreaID();
	local total = 0;
	if (id ~= self.currentZone or force) then
		-- update all pins
		self.currentZone = id;
		self:HideAllPins();
		if (self.data.zones[id]) then
			for q,v in pairs(self.data.zones[id]) do
				if (self:QuestAvailable(q, v)) then	-- if we're not on the quest or have completed it
					if (self.db.profile.display) then
						self:ShowPin(q, v);
					end
					total = total + 1;
					-- self:Debug("Showing pin",q,"on map",id)
				else
					-- self:Debug("Not showing pin",q,"on map",id)
				end
			end
			-- self:Debug("Showing",total,"pins");
		end
		self.mapModule:SetCount(total);
	end
end

function core:ShowPin(id, data)
	local flags, class, race, level, xcoord, ycoord, npc, name = self:GetQuestData(data);
	if (not flags or not tonumber(xcoord) or not tonumber(ycoord)) then
		-- self:Debug("Bad quest data:", id, data, ";;;", flags, level, xcoord, ycoord, npc, name)
		return
	end
	
	if (bit.band(flags, 0x02) > 0) then
		self.data.npcs[tonumber(npc)] = id;
	end
	
	-- check the cache for quest name; scan tooltip if not
	if (not self.cache.quests[id]) then
		self:QuestScan(id);
	end
	if (not self.cache.npcs[npc]) then
		self:NPCScan(npc);
	end
	
	local pin = self:GetPin();
	
	pin:ClearAllPoints();
	pin.questHubberID = id;
	-- pin:SetParent(WorldMapDetailFrame);
	pin:SetFrameLevel(WorldMapPOIFrame:GetFrameLevel()-1);
	pin:SetPoint("CENTER", WorldMapDetailFrame, "TOPLEFT", xcoord/10000*WorldMapDetailFrame:GetWidth(), -ycoord/10000*WorldMapDetailFrame:GetHeight());
	-- pin:SetFrameStrata("FULLSCREEN");
	pin:Show();
	
	self.pinReg[id] = pin;
end

function core:HidePin(id)
	if (self.pinReg[id]) then
		self.pinReg[id]:Hide();
		table.insert(self.frameReg, self.pinReg[id]);
		self.pinReg[id] = nil;
	end
end

function core:HideAllPins()
	for i,v in pairs(self.pinReg) do
		self:HidePin(i);
	end
end

function core:GetPin()
	if (#self.frameReg > 0) then
		return table.remove(self.frameReg);
	end
	
	-- create new
	local pin = CreateFrame("Frame", nil, WorldMapDetailFrame);
	pin.texture = pin:CreateTexture();
	-- 256x64, 8x4
	pin.texture:SetTexture("Interface\\MINIMAP\\ObjectIcons.blp");
	pin.texture:SetTexCoord(0.125, 0.250, 0.250, 0.500);
	pin.texture:SetAllPoints();
	
	pin:SetWidth(16);
	pin:SetHeight(16);
	pin:EnableMouse(true);
	pin:SetScript("OnEnter", function(pin)
		self:ShowTooltip(pin);
	end);
	pin:SetScript("OnLeave", function()
		self:HideTooltip();
	end);
	
	return pin;
end

-- QuestAvailable() returns whether or not the quest is available for pickup right now
function core:QuestAvailable(id, data)
	if data then
		local flags, class, race, level = self:GetQuestData(data);
		class, race, level = tonumber(class), tonumber(race), tonumber(level)
		-- check level requirement
		if (UnitLevel("player") < level) or (UnitLevel("player") > level + 5 and not self:TrackingLow()) then
			-- self:Debug("Hid", id, "because of level restriction")
			return false
		end
		
		-- check race
		local urace = RACE_ID[select(2, UnitRace("player"))]
		if (race > 0 and urace and bit.band(race, urace) == 0) then
			-- self:Debug("Hid", id, "because of race restriction")
			return false
		end			
		
		-- check class
		local uclass = CLASS_ID[select(2, UnitClass("player"))]
		if (class > 0 and uclass and bit.band(class, uclass) == 0) then
			-- self:Debug("Hid", id, "because of class restriction")
			return false
		end		
	end
	
	-- check based on chain
	if (not self.questLog[id] and not self.playerQuests[id] and self.questLink[id]) then
		return not self.playerQuests[self.questLink[id]];
	end
	
	-- now only return true if its not in your logs or completed list
	return not self.questLog[id] and not self.playerQuests[id];
end

-- checks whether the player is tracking low level quests
local trackIndex
function core:TrackingLow()
	if not trackIndex then
		for i=1,GetNumTrackingTypes() do
			local _, texture = GetTrackingInfo(i)
			if texture == "Interface\\Minimap\\Tracking\\TrivialQuests" then
				trackIndex = i
				break
			end
		end
	end
	
	if trackIndex then
		return select(3, GetTrackingInfo(trackIndex))
	end
end

-----------------------------------------------
-- Tooltip management
-----------------------------------------------

function core:GetMousePosition()
	local left, top = WorldMapDetailFrame:GetLeft(), WorldMapDetailFrame:GetTop();
	local width, height = WorldMapDetailFrame:GetWidth(), WorldMapDetailFrame:GetHeight();
	local scale = WorldMapDetailFrame:GetEffectiveScale();

	local x, y = GetCursorPosition();
	local cx = (x/scale - left) / width;
	local cy = (top - y/scale) / height;
	
	return math.min(math.max(cx, 0), 1), math.min(math.max(cy, 0), 1);
end

function core:Distance(x1, y1, x2, y2)
	-- gets distance in map units TODO: use yards?
	local distRatio = WorldMapDetailFrame:GetHeight() / WorldMapDetailFrame:GetWidth();
	
	return math.sqrt( (x1 - x2)^2 + ((y1 - y2)/distRatio)^2 );
end

function core:ShowTooltip(pin)
	self.tooltip:SetOwner(pin, "ANCHOR_RIGHT");
	self.tooltip:ClearLines();
	
	-- find all quests in range of hover
	local mx, my = self:GetMousePosition();
	local npcList = {};
	local npcNames = {};
	for q,v in pairs(self.data.zones[self.currentZone]) do
		if (self:QuestAvailable(q, v)) then
			local _, _, _, _, xcoord, ycoord, npc, npcname = self:GetQuestData(v);
			if (tonumber(xcoord) and tonumber(ycoord)) then
				local dist = self:Distance(mx, my, xcoord/10000, ycoord/10000);
				-- self:Print("d:",dist,v.begin.x,v.begin.y);
				if (dist <= 0.02) then
					if (not npcList[npc]) then
						npcList[npc] = {};
						npcNames[npc] = npcname;
					end
					table.insert(npcList[npc], q);
				end
			end
		end
	end
	
	local first = true;	
	local cZone = self.data.zones[self.currentZone];
	for npc,questList in pairs(npcList) do
		if (not first) then
			self.tooltip:AddLine(" ");
		else
			first = false;
		end
		table.sort(questList);
		for _,q in ipairs(questList) do
			local flags, class, race, level, _, _, _, npcname = self:GetQuestData(cZone[q]);
			local leftStr = format("[%d] %s", level, self.cache.quests[q] or "(Querying server...)")
			local rightStr = "";
			if bit.band(flags, 0x02) > 0 then
				-- rightStr = "|TInterface\\CURSOR\\CURSORICONSNEW:16:16:0:0:128:256:32:63:128:159|t" -- -64:-96
				rightStr = "  |TInterface\\MINIMAP\\TRACKING\\Banker:0|t"-- :21:26:0:0:128:256:37:58:128:154|t" -- -64:-96
			end
			-- self.tooltip:AddDoubleLine(leftStr, rightStr, 255/255, 210/255, 0/255, 1);
			self.tooltip:AddLine(leftStr .. rightStr, 255/255, 210/255, 0/255, 1);
			self.tooltip:SetLastFont(self.tooltip.large);
		end
		-- if the npc name is cached, display it (will be localized); if not, if we are on enUS or enGB, display the hard-coded name as it should be localized; if on a different locale, display the hard-coded name in parenthesis to denote that it has not been discovered yet
		self.tooltip:AddLine(self.cache.npcs[npc] or (IS_ENGLISH and npcNames[npc]) or format("(%s)", npcNames[npc]), 1, 1, 1, 1);
		self.tooltip:SetLastFont(self.tooltip.small);
	end
	
	self.tooltip:Show();
end

function core:HideTooltip()
	self.tooltip:Hide();
end

-----------------------------------------------
-- Quest log management
-----------------------------------------------

function core:QUEST_QUERY_COMPLETE()
	self.playerQuests = GetQuestsCompleted();
	self:UpdatePins(true);
end

function core:QUEST_LOG_UPDATE()
	-- get closed tabs
	local closed = {}
	for i=1,GetNumQuestLogEntries() do
		local title, _, _, _, isHeader, isCollapsed = GetQuestLogTitle(i);
		if (isCollapsed) then
			closed[title] = true
		end
	end
	ExpandQuestHeader(0)
	
	if (not self.questInit) then
		for i=1,GetNumQuestLogEntries() do
			local _, _, _, _, _, _, _, _, id = GetQuestLogTitle(i);
			if (id and id > 0) then
				self.questLog[id] = true;
			end
		end
		self:Debug("Quest Log snapshot saved");
		-- self:UnregisterEvent("QUEST_LOG_UPDATE");
		self.questInit = true;
	else
		-- self:Debug("QUEST_LOG_UPDATE");
		local tempLog = {};
		for i=1,GetNumQuestLogEntries() do
			local name, _, _, _, _, _, _, _, id = GetQuestLogTitle(i);
			-- self:Debug("   ", name, id);
			if (id and id > 0) then
				tempLog[id] = name;
			end
		end
		-- self:Debug("_END_");
		
		for i,v in pairs(self.questLog) do
			if (not tempLog[i]) then
				-- quest removed; hopefully completed
				self:Debug("Quest completed:",i);
				self.questLog[i] = nil;
				tempLog[i] = nil;
				self.playerQuests[i] = true;
				-- force a full update in case any optional pre-reqs should be turned off
				self:UpdatePins(true);
			end
		end
		
		-- merge added
		for i,v in pairs(tempLog) do
			if (not self.questLog[i]) then
				self:Debug("Quest accepted:", v);
				self.questLog[i] = true;
				self:HidePin(i);
			end
		end
	end
	
	-- restore closed tabs
	for i=1,GetNumQuestLogEntries() do
		local title, _, _, _, isHeader = GetQuestLogTitle(i);
		if closed[title] then
			CollapseQuestHeader(i)
		end
	end
end

function core:SetAbandonQuest()
	self.abandonQuest = GetQuestLogSelection();
end

function core:AbandonQuest()
	-- saved quest was abandoned
	-- self:Debug("AbandonQuest");
	local name, _, _, _, _, _, _, _, id = GetQuestLogTitle(self.abandonQuest);
	self:Debug("Quest abandoned:",name);
	self.questLog[id] = nil;
	local quest = self.data.quests[id];
	if (quest and quest.begin and quest.begin.zoneid == self.currentZone) then
		self:ShowPin(id, quest, "begin");
	end
end

function core:Debug(level, ...)
	if (type(level) ~= "number") then
		self:Debug(1, level, ...);
		return;
	end
	if self.db.profile.debug >= level then 
		self:Print(...);
	end
end

local loader  -- timer that runs UpdatePins after everything is loaded
function core:RegisterQuests(zones)
	for i,v in pairs(zones) do
		if (not self.data.zones[i]) then
			self.data.zones[i] = {};
		end
		
		for q,qv in pairs(v) do
			self.data.zones[i][q] = qv;
		end
	end
	self.dataLoaded = true;
	if (not loader) then
		loader = self:ScheduleTimer(function()
			self:UpdatePins(true);
			self:SecureHookScript(GameTooltip, "OnTooltipSetUnit", "CheckNPCTooltip"); -- used for adding notes to mobs' tooltips
		end, 1);
	end
end

-- returns flags, class, race, level, xcoord, ycoord, npc, npcname
function core:GetQuestData(questStr)
	return questStr:match("(%d+):(%d+):(%d+):(%d+):(%d+):(%d+):(%d+):(.+)")
end

function core:SetupMapModule()
	-- frame contains checkbox and text label
	local frame = CreateFrame("Frame", nil, WorldMapDetailFrame);
	frame:SetSize(250, 20);
	frame:SetPoint("TOPLEFT", WorldMapDetailFrame, "TOPLEFT", 5, -2);
	frame:Hide();
	
	frame.check = CreateFrame("CheckButton", nil, frame, "OptionsBaseCheckButtonTemplate");
	frame.check:SetSize(18, 18);
	frame.check:SetPoint("LEFT");
	frame.check:SetChecked(self.db.profile.display);
	frame.check:SetScript("OnClick", function(self)
		if (self:GetChecked()) then
			core.db.profile.display = true
		else
			core.db.profile.display = false
		end
		core:UpdatePins(true)
	end)
	frame.check:SetScript("OnEnter", function(self)
		-- core:Print("Enter");
		core.tooltip:ClearLines();
		core.tooltip:SetOwner(self, "ANCHOR_RIGHT");
		core.tooltip:AddLine(L["Show/Hide all pins"]);
		core.tooltip:Show();
	end);
	frame.check:SetScript("OnLeave", function(self)
		-- core:Print("Leave");
		core.tooltip:Hide();
	end);
	
	local font, size = GameTooltipHeader:GetFont();
	frame.text = frame:CreateFontString();
	frame.text:SetFont(font, size, "OUTLINE");
	frame.text:SetTextColor(255/255, 210/255, 0/255, 1);
	frame.text:SetText("QuestHubber");
	frame.text:SetPoint("LEFT", frame.check, "RIGHT");
	--	<CheckButton name="" inherits="OptionsBaseCheckButtonTemplate" virtual="true">

	frame.note = frame:CreateFontString();
	frame.note:SetFont(font, 10, "NONE");
	frame.note:SetWidth(250);
	frame.note:SetWordWrap(true);
	-- frame.note:SetIndentedWordWrap(0);
	frame.note:SetJustifyH("LEFT");
	frame.note:SetText(L["Note: Your quest completion data has not been retrieved from the server. This could take up to a minute."]);
	frame.note:SetPoint("TOPLEFT", frame.check, "BOTTOMLEFT");
	
	core.mapModule = frame;
	core.mapModule.SetCount = function(self, n)
		-- core:Debug("SetCount", n)
		if (n and n>0) then
			self.text:SetText(format("QuestHubber (%d)", n));
		else
			self.text:SetText(format("QuestHubber"));
		end
	end
end

function core:SetupTooltip()
	self.tooltip = CreateFrame("GameTooltip", "QuestHubberTooltip", UIParent, "GameTooltipTemplate");
	self.tooltip:SetFrameStrata("TOOLTIP");
	self.tooltip.large = QuestHubberTooltipTextLeft1:GetFontObject();
	self.tooltip.small = QuestHubberTooltipTextLeft2:GetFontObject();
	self.tooltip.SetLastFont = function(self, fontObj)
		_G["QuestHubberTooltipTextLeft"..self:NumLines()]:SetFont(fontObj:GetFont());
		-- _G["QuestHubberTooltipTextRight"..self:NumLines()]:SetFont(fontObj:GetFont());
	end
	
	-- if the UI panel disappears (maximized WorldMapFrame) we need to change parents
	self:HookScript(UIParent, "OnHide", function()
		self.tooltip:SetParent(WorldMapFrame);
		self.tooltip:SetFrameStrata("TOOLTIP");
	end);
	self:HookScript(UIParent, "OnShow", function()
		self.tooltip:SetParent(UIParent);
		self.tooltip:SetFrameStrata("TOOLTIP");
	end);
end

function core:CheckNPCTooltip(tooltip)
	if (not UnitIsPlayer("mouseover") or true) then
		-- check if this npc drops a quest item
		local guid = UnitGUID("mouseover");
		if ( guid ) then
			local id = tonumber(guid:sub(7, 10), 16);
			local q = self.data.npcs[id];
			if (q and self:QuestAvailable(q)) then
				local leftStr = format("|TInterface\\MINIMAP\\ObjectIcons:17:9:0:0:256:128:40:56:32:63|t %s", L["Drops an item which starts [%s]"]:format(self.cache.quests[q] or ("(%s)"):format(L["Querying server..."])))
				tooltip:AddLine(leftStr);
				tooltip:Show();
			end
		end
	end
end

function core:MINIMAP_UPDATE_TRACKING()
	self:UpdatePins(true);
end

--[[
local test = {
	[492] = {},
}
local pool = {
	[14265] = "16:3:%04d:%04d:43455:Test Dude",
	[14269] = "16:3:%04d:%04d:43455:Test Dude",
	[14272] = "16:3:%04d:%04d:43455:Test Dude",
	[14273] = "16:3:%04d:%04d:43455:Test Dude",
	[14274] = "16:3:%04d:%04d:43455:Test Dude",
	[14275] = "16:3:%04d:%04d:43455:Test Dude",
	[14276] = "16:3:%04d:%04d:43455:Test Dude",
	[14277] = "16:3:%04d:%04d:43455:Test Dude",
	[14278] = "16:3:%04d:%04d:43455:Test Dude",
	[14279] = "16:3:%04d:%04d:43455:Test Dude",
	[14280] = "16:3:%04d:%04d:43455:Test Dude",
	[14281] = "16:3:%04d:%04d:43455:Test Dude",
	[14283] = "16:3:%04d:%04d:43455:Test Dude",
	[26129] = "16:3:%04d:%04d:43455:Test Dude",
	[14266] = "16:3:%04d:%04d:43455:Test Dude",
	[14285] = "16:3:%04d:%04d:43455:Test Dude",
	[14286] = "16:3:%04d:%04d:43455:Test Dude",
	[14287] = "16:3:%04d:%04d:43455:Test Dude",
	[14288] = "16:3:%04d:%04d:43455:Test Dude",
	[14289] = "16:3:%04d:%04d:43455:Test Dude",
	[14290] = "16:3:%04d:%04d:43455:Test Dude",
	[14291] = "16:3:%04d:%04d:43455:Test Dude",
	[14159] = "16:3:%04d:%04d:43455:Test Dude",
	[14204] = "16:3:%04d:%04d:43455:Test Dude",
	[14220] = "16:3:%04d:%04d:43455:Test Dude",
	[14313] = "16:3:%04d:%04d:43455:Test Dude",
	[14319] = "16:3:%04d:%04d:43455:Test Dude",
	[14320] = "16:3:%04d:%04d:43455:Test Dude",
	[14321] = "16:3:%04d:%04d:43455:Test Dude",
	[14336] = "16:3:%04d:%04d:43455:Test Dude",
	[14347] = "16:3:%04d:%04d:43455:Test Dude",
	[14348] = "16:3:%04d:%04d:43455:Test Dude",
	[14366] = "16:3:%04d:%04d:43455:Test Dude",
	[14375] = "16:3:%04d:%04d:43455:Test Dude",
	[14367] = "16:3:%04d:%04d:43455:Test Dude",
	[14368] = "16:3:%04d:%04d:43455:Test Dude",
	[14369] = "16:3:%04d:%04d:43455:Test Dude",
	[14382] = "16:3:%04d:%04d:43455:Test Dude",
	[14386] = "16:3:%04d:%04d:43455:Test Dude",
	[14395] = "16:3:%04d:%04d:43455:Test Dude",
	[14396] = "16:3:%04d:%04d:43455:Test Dude",
	[14397] = "16:3:%04d:%04d:43455:Test Dude",
	[14398] = "16:3:%04d:%04d:43455:Test Dude",
	[14399] = "16:3:%04d:%04d:43455:Test Dude",
	[14400] = "16:3:%04d:%04d:43455:Test Dude",
	[14401] = "16:3:%04d:%04d:43455:Test Dude",
	[14402] = "16:3:%04d:%04d:43455:Test Dude",
	[14403] = "16:3:%04d:%04d:43455:Test Dude",
	[14404] = "16:3:%04d:%04d:43455:Test Dude",
	[14405] = "16:3:%04d:%04d:43455:Test Dude",
	[14406] = "16:3:%04d:%04d:43455:Test Dude",
	[14412] = "16:3:%04d:%04d:43455:Test Dude",
	[14416] = "16:3:%04d:%04d:43455:Test Dude",
	[14463] = "16:3:%04d:%04d:43455:Test Dude",
	[14465] = "16:3:%04d:%04d:43455:Test Dude",
	[14466] = "16:3:%04d:%04d:43455:Test Dude",
	[24438] = "16:3:%04d:%04d:43455:Test Dude",
	[24468] = "16:3:%04d:%04d:43455:Test Dude",
	[24472] = "16:3:%04d:%04d:43455:Test Dude",
	[24483] = "16:3:%04d:%04d:43455:Test Dude",
	[24484] = "16:3:%04d:%04d:43455:Test Dude",
	[24495] = "16:3:%04d:%04d:43455:Test Dude",
	[24578] = "16:3:%04d:%04d:43455:Test Dude",
	[24501] = "16:3:%04d:%04d:43455:Test Dude",
	[24593] = "16:3:%04d:%04d:43455:Test Dude",
	[24616] = "16:3:%04d:%04d:43455:Test Dude",
	[24627] = "16:3:%04d:%04d:43455:Test Dude",
	[24646] = "16:3:%04d:%04d:43455:Test Dude",
	[24672] = "16:3:%04d:%04d:43455:Test Dude",
	[24673] = "16:3:%04d:%04d:43455:Test Dude",
	[24575] = "16:3:%04d:%04d:43455:Test Dude",
	[24592] = "16:3:%04d:%04d:43455:Test Dude",
	[24675] = "16:3:%04d:%04d:43455:Test Dude",
	[24677] = "16:3:%04d:%04d:43455:Test Dude",
	[24602] = "16:3:%04d:%04d:43455:Test Dude",
	[24628] = "16:3:%04d:%04d:43455:Test Dude",
	[24674] = "16:3:%04d:%04d:43455:Test Dude",
	[24676] = "16:3:%04d:%04d:43455:Test Dude",
	[24678] = "16:3:%04d:%04d:43455:Test Dude",
	[24679] = "16:3:%04d:%04d:43455:Test Dude",
	[24904] = "16:3:%04d:%04d:43455:Test Dude",
	[24680] = "16:3:%04d:%04d:43455:Test Dude",
	[24681] = "16:3:%04d:%04d:43455:Test Dude",
	[26706] = "16:3:%04d:%04d:43455:Test Dude",
}
for x=60,64,1 do
	for y=40,48,1 do
		i,v = next(pool)
		if not i then break end
		test[492][i] = v:format(x*100, y*100)
		pool[i] = nil
	end
end

core:RegisterQuests(test)]]